'' PropCOMM VT100_VGA V2.0b Original By Jeff Ledger

{{
Info            : VT100 Terminal for AVR CP/M
Autor           : Joe G.
Version         : 1.3
Subversion      : 02
Funktion        : Keyboard Driver DE, VGA-640x480 Driver, V24 Driver, VT100 Terminal
Komponenten     : VGA_640
                  Keyb-DE
                  FullDuplexSerial
                  VGA_HiRes_Text_Color

Log

25.03.2013      Start
                V24 Buffer 256 Byte
                Key-DE
26.03.2013      ESC Code Terminal Reset
10.04.2013      new codes
13.04.2013      BS
14.04.2013      ESC [ Y;X H
26.04.2013      arrow Keys, cursor (not) visible, colormanagement
25.05.2013      Bell, CP/M Reset
26.05.2013      all VT100 colors, hot key Alt+i (toggle inverse)
02.01.2014      makes Cursor visible by Reset (F1 und ESC c), correct storage location in the Rx&Tx buffer
03.01.2014      with German letters Ä,Ö,Ü,ä,ö,ü,ß and ESC[H, ESC[f, bug in ESC[M
19.01.2014      second character font
28.01.2014      new VGA Mode 640x480 full color




# Current VT-100 Code list

## ESC sequences:

ESC c     RIS      Reset.
ESC D     IND      Linefeed.
ESC E     NEL      Newline.
ESC H     HTS      Set tab stop at current column.
ESC M     RI       Reverse linefeed.
ESC Z     DECID    DEC private identification. The terminal returns the
                   string  ESC [ ? 6 c, claiming that it is a VT102.
ESC 7     DECSC    Save current state (cursor coordinates,
                   attributes, character sets pointed at by G0, G1).
ESC 8     DECRC    Restore state most recently saved by ESC 7.
ESC [     CSI      Control sequence introducer
ESC # 8   DECALN   DEC screen alignment test - fill screen with E's.
ESC # 9            showfont  (experimantal)
ESC (              Start sequence defining G0 character set
ESC ( B               Select default (ISO 8859-1 mapping)
ESC ( 0               Select VT100 graphics mapping
ESC )              Start sequence defining G1
                   (followed by one of B, 0 as above).
ESC >     DECPNM   Set numeric keypad mode
ESC =     DECPAM   Set application keypad mode


## ECMA-48 CSI sequences

CSI (or ESC [) is followed by a sequence of parameters, at most NPAR
(16), that are decimal numbers separated by semicolons.  An empty or
absent parameter is taken to be 0.  The sequence of parameters may be
preceded by a single question mark.

The action of a CSI sequence is determined by its final character.

@   ICH       Insert the indicated # of blank characters.
A   CUU       Move cursor up the indicated # of rows.
B   CUD       Move cursor down the indicated # of rows.
C   CUF       Move cursor right the indicated # of columns.
D   CUB       Move cursor left the indicated # of columns.
H   CUP       Move cursor to the indicated row, column (origin at 1,1).
J   ED        Erase display (default: from cursor to end of display).
              ESC [ 1 J: erase from start to cursor.
              ESC [ 2 J: erase whole display.
K   EL        Erase line (default: from cursor to end of line).
              ESC [ 1 K: erase from start of line to cursor.
              ESC [ 2 K: erase whole line.
L   IL        Insert the indicated # of blank lines.
M   DL        Delete the indicated # of lines.
P   DCH       Delete the indicated # of characters on current line.
X   ECH       Erase the indicated # of characters on current line.
c   DA        Answer ESC [ ? 6 c: "I am a VT102".
f   HVP       Move cursor to the indicated row, column.
g   TBC       Without parameter: clear tab stop at current position.
              ESC [ 3 g: delete all tab stops.
h   SM        Set Mode (see below).
l   RM        Reset Mode (see below).
m   SGR       Set attributes (see below).
n   DSR       Status report (see below).
q   DECLL     Set keyboard LEDs.
              ESC [ 0 q: clear all LEDs
              ESC [ 1 q: set Scroll Lock LED
              ESC [ 2 q: set Num Lock LED
              ESC [ 3 q: set Caps Lock LED
r   DECSTBM   Set scrolling region; parameters are top and bottom row.



## ECMA-48 Set Graphics Rendition

The ECMA-48 SGR sequence ESC [ parameters m sets display attributes.
Several attributes can be set in the same sequence, separated by
semicolons.  An empty parameter (between semicolons or string
initiator or terminator) is interpreted as a zero.

param   result
0       reset all attributes to their defaults
1       set bold
2       set half-bright (simulated with color on a color display)
4       set underscore (simulated with color on a color display)
        (the colors used to simulate dim or underline are set
        using ESC ] ...)
5       set blink
7       set reverse video
10      Primary(default) font (experimantal)
11      1st alternate font (experimantal)
12      2nd alternate font (experimantal)
21      set normal intensity (ECMA-48 says "doubly underlined")
22      set normal intensity
24      underline off
25      blink off
27      reverse video off
30      set black foreground
31      set red foreground
32      set green foreground
33      set brown (yellow) foreground
34      set blue foreground
35      set magenta foreground
36      set cyan foreground
37      set white foreground
38      set underscore on, set default foreground color
39      set underscore off, set default foreground color
40      set black background
41      set red background
42      set green background
43      set brown (yellow) background
44      set blue background
45      set magenta background
46      set cyan background
47      set white background
49      set default background color


## ECMA-48 Mode Switches

Control sequence 'ESC [ Ps ; ... h' sets one or more modes specified
by selective parameters (Ps) in the parameter string.
Control sequence 'ESC [ Ps ; ... l' resets one or more modes specified
by selective parameters (Ps) in the parameter string.

Ps   Mnemonic   Description

3    DECCRM     (default off): Display control chars.
4    DECIM      (default off): Set insert mode.
20   LF/NL      (default off): Automatically follow echo of LF, VT or FF
                with CR.

## DEC Private Mode (DECSET/DECRST) sequences

'ESC [ ? Ps ; ... h' sets one or more private modes.
'ESC [ ? Ps ; ... l' resets one or more private modes.

Ps   Mnemonic   Description
3    DECCOLM    (default off = 80 columns): 80/132 col mode switch.
                NOTE: When you change the # of columns per line, the screen
                is erased! This also sets the scrolling region for full screen.
5    DECSCNM    (default off): Set reverse-video mode.
6    DECOM      (default off): When set, cursor addressing is relative
                to the upper left corner of the scrolling region.
7    DECAWM     (default on): Set autowrap on.  In this mode, a graphic
                character emitted after column 80 (or column 132 of DECCOLM is
                on) forces a wrap to the beginning of the following line
                first.
25   DECTECM    (default on): Make cursor visible.


## ECMA-48 Status Report Commands

ESC [ 5 n       Device status report (DSR):
                Answer is ESC [ 0 n (Terminal OK).
                       or ESC [ 3 n (Terminal not OK)
ESC [ 6 n       Cursor position report (CPR): Answer is ESC [ y ; x R, where
                x,y is the cursor location.

## Linux Console Private CSI Sequences
(not implemented yet)

The following sequences are neither ECMA-48 nor native VT102.  They
are native to the Linux console driver.  Colors are in SGR
parameters: 0 = black, 1 = red, 2 = green, 3 = brown, 4 = blue, 5 =
magenta, 6 = cyan, 7 = white.

ESC [ 1 ; n ]       Set color n as the underline color
ESC [ 2 ; n ]       Set color n as the dim color
ESC [ 8 ]           Make the current color pair the default attributes.


}}

{{
Terminal Input Parser
see http://www.vt100.net/emu/dec_ansi_parser


State           Event           Next State      Action
--------------------------------------------------------------
'anywhere'      00-17,19,1C-1F  -               execute
                18, 1A          ground          execute
                1B              escape          clear_esc
--------------------------------------------------------------
ground          20-7F           -               print
--------------------------------------------------------------
escape          7F              -               ignore
                20-2F           esc_interm      collect
                5B              csi_entry       clear_csi
                30-5A,5C-7E     ground          esc_dispatch
--------------------------------------------------------------
esc_interm      7F              -               ignore
                20-2F           -               collect
                30-7E           ground          esc_dispatch
--------------------------------------------------------------
csi_entry       7F              -               ignore
                20-2F           csi_interm      collect
                3A              csi_ignore      -
                30-39,3B        csi_param       csi_param
                3C-3F           csi_param       collect
                40-7E           ground          csi_dispatch
--------------------------------------------------------------
csi_interm      7F              -               ignore
                20-2F           -               collect
                30-3F           csi_ignore      -
                40-7E           ground          csi_dispatch
--------------------------------------------------------------
csi_ignore      20-3F,7F        -               ignore
                40-7E           ground          -
--------------------------------------------------------------
csi_param       7F              -               ignore
                20-2F           csi_interm      collect
                30-39,3B        -               csi_param
                3A,3C-3F        csi_ignore      -
                40-7E           ground          csi_dispatch
--------------------------------------------------------------

}}

CON
  'clock configuration
  _clkmode  = xtal1 + pll16x
  _xinfreq  = 5_000_000         '5 MHz Quarz AVR CP/M

  'Hardware Variant
  #0, AVRCPM, VT100

  'pin configuration
  KEY_DATA    = 26                'Keyboard Data Pin
  KEY_CLK     = 27                'Keyboard CLK Pin
  VIDEO       = 16                'Video output Pin(s)
  TXD         = 30                'Serial Transmit
  RXD         = 31                'Serial Receive

  'Only on VT100:
  PORTSWITCH  = 6                'Communication Port Switch Pin
  BELL_VT100  = 7                 'Beep

  'Only on AVR-CP/M
  BELL_AVRCPM = 1                 'Beep
  CPM_RESET   = 0                 'CP/M Reset

' ISO-6429 color numbers
  #0, black,    red,  green,  yellow,  blue,  magenta,  cyan, gray
     darkgray, lred, lgreen, lyellow, lblue, lmagenta, lcyan, white

' Text attributes
  ATTR_BOLD       = text#ATTR_BOLD
  ATTR_INVERS     = text#ATTR_INVERS
  ATTR_UNDERLINE  = text#ATTR_UNDERLINE


'Control Characters Recognized by VT102

  NUL           = $00           'Ignored when received (not stored in input buffer) and used as a fill character.
  ETX           = $03           'Can be selected as a half-duplex turnaround character.
  EOT           = $04           'Can be selected as a disconnect character or half-duplex turnaround character. When used as a turnaround character, the disconnect character is DLE-EOT.
  ENQ           = $05           'Transmits answerback message.
  BEL           = $07           'Generates bell tone.
  BS            = $08           'Moves cursor to the left one character position; if cursor is at left margin, no action occurs.
  HT            = $09           'Moves cursor to next tab stop, or to right margin if there are no more tab stops.
  LF            = $0A           'Causes a linefeed or a new line operation. (See Linefeed/New Line). Also causes printing if auto print operation is selected.
  VT            = $0B           'Processed as LF.
  FF            = $0C           'Processed as LF. FF can also be selected as a half-duplex turnaround character.
  CR            = $0D           'Moves cursor to left margin on current line. CR can also be selected as a half-duplex turnaround character.
  SO            = $0E           'Selects G1 character set designated by a select character set sequence.
  SI            = $0F           'Selects G0 character set designated by a select character set sequence.
  DC1           = $11           'Processed as XON. DC1 causes terminal to continue transmitting characters.
  DC3           = $13           'Processed as XOFF. DC3 causes terminal to stop transmitting all characters except XOFF and XON. DC3 can also be selected as a half-duplex turnaround character.
  CAN           = $18           'If received during an escape or control sequence, cancels the sequence and displays substitution character ([]).
  SUBST         = $1A           'Processed as CAN.
  ESC           = $1B           'Processed as a sequence introducer.
  DEL           = $7F

  bspKey            = $C8
  delKey            = $BA
  ArrowTop          = $C2
  ArrowDown         = $C3
  ArrowLeft         = $C0
  ArrowRight        = $C1

  escape            = $CB
  backspace         = 8
  fReturn           = 13
  fLinefeed         = 10
  fEof              = -1
  CSI_PAR_MAX        = 10        'Maximum number of escape sequence parameters

  ' Input parser states
  #0, S_GROUND
  S_ESCAPE
  S_ESCAPE_INTERMEDIATE
  S_CSI_ENTRY
  S_CSI_INTERMEDIATE
  S_CSI_IGNORE
  S_CSI_PARAM

  'Communication Port
  #0, comV24, comUSB

DAT
  strVersion byte "VT100 Terminal V1.7 test-multifont ",0
  strNewline byte $0D,$0A,0
  strV24 byte "V24",0
  strUSB byte "USB",0

  BaudTab long 300, 1200, 2400, 4800, 9600, 19200, 38400, 57600, 115200

  PortTab word @strV24, @strUSB

Var
  word  attr_color              'low byte: color; high byte: other attributes (underline, invers)
  byte  c_mode                  'cursor mode
  byte  menu_flag               'menu flag
  byte  baud                    'baud rate (index)
  byte  fg_color                'foreground color
  byte  bg_color                'background color
  byte  comport                 'I/O port
  byte  variant
  byte  bell_pin
  byte  font

  ' Input parser state variables
  long   intermediate_char
  long   csi_par_cnt
  long   csi_par[CSI_PAR_MAX]

'Temp starage
  long  cfg[config#CfgSize]     'configuration storage

OBJ
  text   : "VGA_640"                   ' Full Color VGA-640x480 Driver
  key    : "keyb-DE"                   ' new Keyboard Driver (german)
  ser    : "FullDuplexSerial"          ' Full Duplex Serial Controller
  config : "eeprom"                    ' I2C EEPROM access

PUB start

  init                                                  'initialization
  vt_reset                                              '(pre)set parameters
  titlescreen                                           'print startmessage

{
  repeat
    result := key.getkey
    Text.hex(result,3)
    Text.str(string(" : "))
    text.print(result)
    text.print($0d)
    ser.tx(result)
}
    terminal                                            'run VT100 Terminal Programm


PRI init

  dira[PORTSWITCH] := 0                                 'test portswitch pin
  if (variant := ina[PORTSWITCH]) == AVRCPM
    dira[CPM_RESET] := 0                                'input (open drain)
    bell_pin := BELL_AVRCPM
  else
    bell_pin := BELL_VT100

  dira[bell_pin] := 1                                   'output

  config.initialize                                     'initialize EEPROM

  if config.readCfg(@cfg) == 0                          'read EEPROM, is config zero, set default config
     cfg[0] := 8                                        'baud
     cfg[1] := green                                    'foreground color
     cfg[2] := black                                    'background color
     cfg[3] := 6                                        'cursor mode
     cfg[4] := comUSB                                   'switch port, EN is high (select FTDI, deselect MAX3241)
     cfg[5] := 1                                        'font


  baud := cfg[0]                                        'baud rate is default (115200)
  fg_color := cfg[1]                                    'foreground color is default (dgreen)
  bg_color := cfg[2]                                    'background color is default (black)
  c_mode := cfg[3]                                      'cursor mode is default 6 (underscore, cursor on, blink slow)
  comport := cfg[4]                                     'port 1 is default (USB)
  font := cfg[5]

  outa[PORTSWITCH] := comport                           'set portswitch to configured state
  dira[PORTSWITCH] := 1                                 '  ..and output it
  attr_color := bg_color<<4 + fg_color                  'default color

  text.start(VIDEO)                                     'start terminal screen
  key.start(KEY_DATA,KEY_CLK)                           'start new keyboard driver
  ser.start(RXD,TXD,0,BaudTab[baud])                    'start host communication
  text.cursor_set(c_mode)                               'set cursor mode
  text.set_font(font)

  menu_flag := 0                                        'no help menu



PRI titlescreen
  beep
  text.str(@strVersion)
  text.str(@strNewLine)
  text.dec(BaudTab[baud])
  text.str(string(",8,N,1"))
  if variant == VT100
    text.str(string(" | Port: "))
    text.str(@@PortTab[comport])
  text.str(@strNewLine)
  text.str(string("Terminal Ready",$0D,$0A))

PRI vt_reset
  text.cmd(text#C_DECAWM+1)                             'Autowrap on
  text.cmd(text#C_DECTECM+1)                            'makes cursor visible
  attr_color := bg_color<<4 + fg_color                  'default color
  text.attr_color(attr_color)
  text.ed(2)
  text.cursloc(0,0)

PRI beep
' TODO: beep in background (extra cog or timer)

  outa[bell_pin] := 1                           'set bell high
  waitcnt(clkfreq/8 + cnt)                      'wait a moment (125ms)
  outa[bell_pin] := 0                           'set bell low


PRI do_cpm_reset

  outa[CPM_RESET] := 0                          'set CP/M reset low (just in case)
  dira[CPM_RESET] := 1                          'output (open drain active)
  waitcnt(clkfreq/128 + cnt)                    'wait a moment (7.8ms)
  dira[CPM_RESET] := 0                          'input (inactive)


PRI terminal | ch, vt100_state

  vt100_state := S_GROUND

  repeat
    ch := key.key

    if ch > 0

      if ch > 608 and ch < 635                 'CRTL code
        ch:=ch-608                            'redefine as CTRL+char

      if ch > 451 and ch < 477                 'Ä,Ü,Ö
        ch:=ch-256                            'redefine as ASCII Letter Ä,Ü,Ö

      if ch == 208                                  'if F1
        ch := 0
        helpmenu(1)                                 'print Help Menu
'        vt_reset
'        titlescreen

      if ch == 209                                  'if F2 (Baud Rate)
        ch := $0                                    'code zero
        baud++                                      'increment symbol rate
        if baud == 9
          baud := 0
        helpmenu(0)
        ser.start(RXD,TXD,0,BaudTab[baud])          'restart V24 with new symbol rate

      if ch == 210                                  'if F3 (Fore Color)
        ch := $0                                    'code zero
        fg_color := ++fg_color // 16                'increment foreground colur
        attr_color := attr_color & $FFF0 | fg_color
        text.attr_color(attr_color)                           'set textcolor
        helpmenu(0)

      if ch == 211                                  'if F4 (Back Color)
        ch := $0                                    'code zero
        bg_color := ++bg_color // 16                'increment background colur
        attr_color := attr_color & $FF0F | bg_color<<4
        text.attr_color(attr_color)                           'set textcolor
        helpmenu(0)

      if ch == 212                                  'if F5 (Port)
        if variant == VT100
          ch := $0                                    'code zero
          comport ^= 1                                'toggle comport
          outa[PORTSWITCH] := comport                 'toggle portswitch pin
          ser.rxflush                                 'clear Rx buffer
          helpmenu(0)                                 'visible comport switch

      if ch == 213                                  'if F6 (Cursor)
        ch := $0                                    'code zero
        c_mode := ++c_mode // 8                     'increment cursor mode
        text.cursor_set(c_mode)                     'set corsor mode
        helpmenu(0)                                 'visible c_mode

      if ch == 214                                  'if F7 (Font)
        ch := $0                                    'code zero
        font := ++font // 3                         'increment Font #
        text.set_font(font)
        helpmenu(0)                                 'visible c_mode

      if ch == 217                                  'if F10 (Save+Exit)
        ch := $0                                    'code zero
        helpmenu(-1)                                'hide Help Menu

        cfg[0] := baud                              'baud
        cfg[1] := fg_color                          'foreground color
        cfg[2] := bg_color                          'background color
        cfg[3] := c_mode                            'cursor mode
        cfg[4] := comport                           'switch comport
        cfg[5] := font
        config.writeCfg(@cfg)                       'write EEPROM with configdata
        vt_reset
        titlescreen                                 'clear menu

      if ch == $0469                                'Alt + i
        ch := 0
        attr_color := bg_color<<4 + fg_color  'set default color
        text.attr_color(attr_color)
        beep                                        'beep

      if ch == $06BA or ch == $06C9                 'Ctrl-Alt-Delete
        if variant == AVRCPM
          ch := 0
          do_cpm_reset                                'CP/M Reset

      case ch
        bspKey:
          ser.tx($08)                           'send BS

        delKey:                                 'send DEL
          ser.tx($7F)

        ArrowTop:                               'send ESC [ A
          ser.tx($1B)
          ser.tx($5B)
          ser.tx($41)

        ArrowDown:
          ser.tx($1B)                           'send ESC [ B
          ser.tx($5B)
          ser.tx($42)

        ArrowRight:                             'send ESC [ C
          ser.tx($1B)
          ser.tx($5B)
          ser.tx($43)

        ArrowLeft:                              'send ESC [ D
          ser.tx($1B)
          ser.tx($5B)
          ser.tx($44)

        0:                                      'send nothing

        OTHER:                                  'send code
          ser.tx(ch)

    ch := ser.rxcheck                         'receive char, ignore NUL

      if ch < $20
        'Process control characters immediately. (everywhere pseudostate)
        case ch

          'ENQ:                               'ENQ: Transmits answerback message.
            'TODO
          BEL:                                'BEL: Generates bell tone.
            beep
          BS:                                 'BS: Moves cursor to the left one character position; if cursor is at left margin, no action occurs.
            text.cmd(text#C_CUB)
                                              'HT: Moves cursor to next tab stop, or to right margin if there are no more tab stops.
          HT:
            text.cmd(text#C_HT)               'CR: Moves cursor to left margin on current line.
          CR:
            text.cmd(text#C_CR)
                                              'LF: Causes a linefeed or a new line operation. (See Linefeed/New Line).
                                              'VT: Processed as LF.
                                              'FF: Processed as LF.
          LF, VT, FF:
            text.cmd(text#C_IND)

          'SO                                 'Selects G1 character set designated by a select character set sequence.
          'SI                                 'Selects G0 character set designated by a select character set sequence.
          'DC1                                'Processed as XON. DC1 causes terminal to continue transmitting characters.
          'DC3                                'Processed as XOFF. DC3 causes terminal to stop transmitting all characters except XOFF and XON. DC3 can also be selected as a half-duplex turnaround character.
            'TODO

                                              'CAN: If received during an escape or control sequence, cancels the sequence and displays substitution character ([]).
                                              'SUB: Processed as CAN.
          CAN, SUBST:
            'TODO: text.print(**subst char**)
            vt100_state := S_GROUND

          ESC:                                'ESC: Processed as a sequence introducer.
            'clear_esc
            intermediate_char := 0
            vt100_state := S_ESCAPE

      else

        case vt100_state
          S_GROUND:
              text.print(ch)

          S_ESCAPE:
              case ch
                DEL:
                    'ignore

                $20..$2F:                      '" ".."/"
                    collect(ch)
                    vt100_state := S_ESCAPE_INTERMEDIATE

                $5B:                           '"["
                    'clear csi
                    csi_par_cnt := 0
                    csi_par[0] := 0
                    vt100_state := S_CSI_ENTRY

                other:
                    esc_dispatch(ch)
                    vt100_state := S_GROUND

          S_ESCAPE_INTERMEDIATE:
              case ch
                DEL:
                    'ignore

                $20..$2F:                      '" ".."/"
                    collect(ch)
                    vt100_state := S_CSI_INTERMEDIATE

                other:
                    esc_dispatch(ch)
                    vt100_state := S_GROUND

          S_CSI_ENTRY:
              case ch
                DEL:
                    'ignore

                $20..$2F:                      '" ".."/"
                    collect(ch)

                $3A:                           '":"
                    vt100_state := S_CSI_IGNORE

                $30..$39,$3B:                  '"0".."9",";"
                    param(ch)
                    vt100_state := S_CSI_PARAM

                $3C..$3F:                      '"<".."?"
                    collect(ch)
                    vt100_state := S_CSI_PARAM

                other:
                    csi_dispatch(ch)
                    vt100_state := S_GROUND

          S_CSI_INTERMEDIATE:
              case ch
                DEL:
                    'ignore

                $20..$2F:                      '" ".."/"
                    collect(ch)

                $30..$3F:                      '"0".."?"
                    vt100_state := S_CSI_IGNORE

                other:
                    csi_dispatch(ch)
                    vt100_state := S_GROUND

          S_CSI_IGNORE:
              case ch
                $20..$3F,DEL:                  '" ".."?",DEL
                    'ignore

                other:
                    vt100_state := S_GROUND

          S_CSI_PARAM:
              case ch
                DEL:
                    'ignore

                $20..$2F:                      '" ".."/"
                    collect(ch)
                    vt100_state := S_CSI_INTERMEDIATE

                $3A,$3C..$3F:                  '":","<".."?"
                    vt100_state := S_CSI_IGNORE

                $30..$39,$3B:                  '"0".."9",";"
                    param(ch)

                other:
                    csi_dispatch(ch)
                    vt100_state := S_GROUND



PRI collect (ch)
   if intermediate_char == 0                    'Store only one (the first) intermediate char
      intermediate_char := ch

PRI param (ch)
   if csi_par_cnt < CSI_PAR_MAX
      if ch == ";"
         csi_par_cnt++
         if csi_par_cnt < CSI_PAR_MAX
            csi_par[csi_par_cnt] := 0
      else
         csi_par[csi_par_cnt] := csi_par[csi_par_cnt] * 10 + ch - "0"

PRI get_param (index)
  if index > csi_par_cnt
    return 0
  return csi_par[index]


PRI esc_dispatch (ch)

  case intermediate_char
    0:
      case ch
        "c":                                    'RIS    Reset to Initial State (Power Up)
          vt_reset
          titlescreen
        "D":                                    'IND    Linefeed.
          text.cmd(text#C_IND)
        "E":                                    'NEL    Newline.
          text.cmd(text#C_CR)
          text.cmd(text#C_IND)
        "H":                                    'HTS    Set tab stop at current column.
          'TODO?
        "M":                                    'RI     Reverse linefeed. (Curser up)
          text.cmd(text#C_RI)
        "Z":                                    'DECID  DEC private identification. The terminal returns  ESC [ ? 6 c, claiming that it is a VT102.
          tx_csi(string("?6c"))
        "7":                                    'DECSC  Save current state (cursor coordinates, attributes, character sets pointed at by G0, G1).
          text.cmd(text#C_DECSC)
        "8":                                    'DECRC  Restore state most recently saved by ESC 7.
          text.cmd(text#C_DECRC)

    "?":                                        'Private functions
      'none (yet)
    other:
      case ch
        "8":                                    'DECALN   DEC screen alignment test - fill screen with E's.
          if intermediate_char == "#"
              Text.decaln
      case ch
        "9":                                    '
           if intermediate_char == "#"
              Text.showfont


PRI csi_dispatch (ch) | i

  case intermediate_char
    0:
      case ch
        "@":                                    'ICH    Insert character(s) (VT220)
          text.ich(get_param(0))
        "A":                                    'CUU    Move cursor up
          repeat (get_param(0) #>1)
            text.cmd(text#C_CUU)

        "B":                                    'CUD    Move cursor down
          repeat (get_param(0) #>1)
            text.cmd(text#C_CUD)

        "C":                                    'CUF    Move cursor right
          repeat (get_param(0) #>1)
            text.cmd(text#C_CUF)

        "D":                                    'CUB    Move cursor left
          repeat (get_param(0) #>1)
            text.cmd(text#C_CUB)

        "H", "f":                               'CUP,HVP Move cursor to row, column (origin at 1,1).
          text.cursloc((get_param(1) #>1) - 1, (get_param(0) #>1) - 1)

        "J":                                    'ED     Erase Display
          text.ed(get_param(0))

        "K":                                    'EL     Erase in line
          text.el(get_param(0))

        "L":                                    'IL     Insert line(s)
          repeat (get_param(0) #>1)
            text.cmd(text#C_IL)

        "M":                                    'DL     Delete line(s)
          repeat (get_param(0) #>1)
            text.cmd(text#C_DL)

        "P":                                    'DCH    Delete character(s)
          text.dch(get_param(0))

        "X":                                    'ECH    Erase character(s) (VT220)
          text.ech(get_param(0))

        "m":                                    'SGR    Set character attributes
          repeat i from 0 to csi_par_cnt
            case result := get_param(i)
              0:                                      'All attributes off
                attr_color := bg_color<<4 | fg_color   '(default color)
              1:                                      'bold (increased intensity)
                attr_color |= ATTR_BOLD               'set bit for high color
              2:                                      'half-bright (simulated with color)
              4:                                      'underline (simulated with color)
                attr_color |= ATTR_UNDERLINE          'set bit for underline

              5:                                      'blink
              7:                                      'negative (reverse) image
                attr_color |= ATTR_INVERS
              10..12:                                 'font
                font := result
              22:                                     'bold off (normal intensity)
                attr_color &= ! ATTR_BOLD
              24:                                     'not underlined
                attr_color &= ! ATTR_UNDERLINE        'underline off
              25:                                     'not blinking
              27:                                     'positive image
                attr_color &= ! ATTR_INVERS
              30..37:                                 'fg (character) color
                attr_color := (attr_color & %11111111_11111000) | (result - 30)
              39:                                     'default fg color
                'TODO: attr_color := (attr_color & %11111111_11111000) | fg_color
              40..47:                                 'bg color
                attr_color := ((result - 40) << 4) | (attr_color & %11111111_00001111)
              49:                                     'default bg color
                'TODO: attr_color := (bg_color << 4) | (attr_color & %11111111_00001111)

          text.attr_color(attr_color)
          if font => 10
            text.set_font(font -= 10)

        "c":                                    'DA     Answer ESC [ ? 6 c: "I am a VT102".
          tx_csi(string("?6c"))

        "g":                                    'TBC    Without parameter: clear tab stop at current position.
          'TODO or not TODO?                    '       ESC [ 3 g: delete all tab stops.

        "h":                                    'SM     'Set Mode
          repeat i from 0 to csi_par_cnt
            case get_param(i)
              3:                                'DECCRM Display control chars. (default off)
                ' TODO?
              4:                                'DECIM  Set insert mode. (default off)
                ' TODO?
              20:                               'LF/NL  Automatically follow echo of LF, VT or FF with CR. (default off)
                ' TODO:

        "l":                                    'RM     'Reset Mode
          repeat i from 0 to csi_par_cnt
            case get_param(i)
              3:                                'DECCRM Display control chars. (default off)
                ' TODO?
              4:                                'DECIM  Set insert mode. (default off)
                ' TODO?
              20:                               'LF/NL  Automatically follow echo of LF, VT or FF with CR. (default off)
                ' TODO:

        "r":                                    'DECSTBM Set scrolling region; parameters are top and bottom row.
          'TODO or not TODO?

        "n":                                    'DSR    Status report
           repeat i from 0 to csi_par_cnt
              case get_param(i)
                5:                              'DSR    Device status report: Answer is ESC [ 0 n (Terminal OK).
                  tx_csi(string("0n"))
                6:                              'CPR    Cursor position report: Answer is ESC [ y ; x R
                  'repeat 1
                  waitcnt(10*clkfreq/BaudTab[baud] + cnt)          'debug wait 1 char time
                  result := text.get_cursor
                  ser.tx(ESC)
                  ser.tx("[")
                  ser.dec(result >> 16 + 1)
                  ser.tx(";")
                  ser.dec(result & $FF + 1)
                  ser.tx("R")
        "]":                                    'From Linux console
          case get_param(0)
            1:                                  'Set color n as the underline color
              'TODO
            2:                                  'Set color n as the dim color
              'TODO
            8:                                  'Make the current color pair the default attributes.
              'TODO

    "?":                                        'Private mode
      case ch
        "h":                                    'DECSET Set Mode
          repeat i from 0 to csi_par_cnt
            case get_param(i)
              3:                                'DECCOLM Number of columns (default off = 80 cols)
                text.ed(2)                      'Not supported, but clear the sceen anyway
                text.cursloc(0,0)
              5:                                'DECSCNM Set reverse-video mode. (default off)
              6:                                'DECOM   Origin Mode; curser pos. in scrolling region. (default off)
              7:                                'DECAWM  Set autowrap on/off. (default on)
                text.cmd(text#C_DECAWM+1)
              25:                               'DECTECM Cursor visible
                text.cmd(text#C_DECTECM+1)

        "l":                                    'DECRST Reset Mode
          repeat i from 0 to csi_par_cnt
            case get_param(i)
              3:                                'DECCOLM Number of columns (default off = 80 cols)
                text.ed(2)                      'Not supported, but clear the sceen anyway
                text.cursloc(0,0)
              5:                                'DECSCNM Set reverse-video mode. (default off)
              6:                                'DECOM   Origin Mode; curser pos. in scrolling region. (default off)
              7:                                'DECAWM  Set autowrap on/off. (default on)
                text.cmd(text#C_DECAWM+0)
              25:                               'DECTECM Cursor not visible
                text.cmd(text#C_DECTECM+0)

    other:
      '


PRI tx_csi(seq)

  'repeat 5
  waitcnt(10*clkfreq/BaudTab[baud] + cnt)               'debug wait 1 char time
  ser.tx($1B)                                           'send ESC [
  ser.tx($5B)
  ser.str(seq)

PRI helpmenu(flag) |  cursorpos

  cursorpos := text.get_cursor                                'save cursor position

  case flag
    1:                                                        'open help menu
      if not menu_flag
        text.attr_color(cyan << 4 + black)
        text.cursloc((text#cols-strsize(@strVersion))>>1, 27)
        text.el(2)
        text.str(@strVersion)
        text.cursloc(8, 29)
        text.el(2)
        text.str(string("F2               F3/F4                       F6        F7         F10"))
        if variant == VT100
          text.cursloc(41, 29)
          text.str(string("F5"))
        text.attr_color(black << 4 + cyan)
        text.cursloc(0, 28)
        text.str(string("Baud Rate:          Fore/Back Color              Cursor:     Font:     Save+Exit"))
        if variant == VT100
          text.cursloc(37, 28)
          text.str(string("Port:"))
        menu_flag := true

    0:                                                        'update menu

    -1:                                                       'close menu
      if menu_flag
        text.cursloc(0, 27)
        text.attr_color(bg_color << 4 + fg_color)
        text.ed(0)
        menu_flag := false

  if menu_flag
    text.attr_color(black << 4 + cyan)
    text.cursloc(10, 28)
    text.decw(BaudTab[baud], 6, " ")
    if variant == VT100
      text.cursloc(43, 28)
      text.str(@@PortTab[comport])
    text.cursloc(57, 28)
    text.dec(c_mode)
    text.cursloc(67, 28)
    text.dec(font)
    text.cursloc(30, 28)
    text.attr_color(bg_color << 4 + fg_color)
    text.decw(fg_color, 2,"0")
    text.str(string("/"))
    text.decw(bg_color, 2,"0")
    'text.str(string("Color"))

  text.cursloc(cursorpos & $FFFF, cursorpos >> 16)            'restore old cursor position
  text.attr_color(attr_color)                                 'resore old screen color and attributes

{{

┌──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────┐
│                                                   TERMS OF USE: MIT License                                                  │
├──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────┤
│Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation    │
│files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    │
│modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software│
│is furnished to do so, subject to the following conditions:                                                                   │
│                                                                                                                              │
│The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.│
│                                                                                                                              │
│THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          │
│WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         │
│COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   │
│ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         │
└──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────┘
}}
